# 
# Adaptive Nelder-Mead Method
# ver.1.1b
#
# History:
# [1] Nelder-Mead Method
# https://nykergoto.hatenablog.jp/entry/2019/11/06/Nelder_Mead_Method_%E3%82%92_python_%E3%81%A7%E5%AE%9F%E8%A3%85
# [2] Adaptive Nelder-Mead Method
# 2020-06-01 ver.0.9 first version
# http://ar.nyx.link/
# some codes are added and modified
# so that the original codes are useful, and farther, the Adaptive Nelder-Mead Method is added
# 2020-06-15 ver.1.0 bug fix and added some check code if the result is reasonable
# 2020-09-15 ver.1.1 suport more flexible argument x0
# 2020-09-21 ver.1.1a bug fix for one-dimensional case
# 2020-10-14 ver.1.1b bug fix (note by "### Fix"), allowed int for the lists in x0, changed a misleading name
#	and suppressed excessive warnings
#
# Kenar (Kenji Arisawa)
# email: karisawa@gmail.com


import numpy as np

debug = True

def dprint(*t,**k):
  if debug: print("#DBG",*t,**k)

# Standard Nelder Mead Method without inside contraction
# x0 is an initial value, a list of points
def snmsearch(objective, x0, err=1e-5,**k): # Kenar
	# a point is a list of n floats or n ints
	# x0 may be a point or a list of points 
	# the "list" may be a tuple,np.ndarray
	if isinstance(x0[0],(list,tuple,np.ndarray)):
		eval_points = np.array(x0,dtype=float)		### Fix
	else:
		init_point =  np.array(x0,dtype=float)		### Fix
		scaling_factor = 1.
		if 'scale' in k:
			scaling_factor = k['scale']
		eval_points = np.vstack([init_point + np.eye(len(x0)) * scaling_factor, init_point])
	alpha = 1.
	gamma = 2.
	rho = .5
	sigma = .5	
	center = np.mean(eval_points, axis=0)
	# loop while radius of the simplex is larger than err
	while np.max(abs(eval_points - center)) > err: ### Fix,Kenar
	    # 目的関数で評価して並び替え
	    objective_values_by_point = objective(np.transpose(eval_points)) # Kenar
	    orders = np.argsort(objective_values_by_point)
	    f_values = objective_values_by_point[orders]
	    eval_points = eval_points[orders]
	
	    # 最悪点以外で重心を計算
	    centroid = np.mean(eval_points[:-1], axis=0)
	    p_worst = eval_points[-1]
	
	    # 反射点 (最悪点から重心の対象な点) を計算 + そのときの目的関数を評価
	    p_reflect = centroid + alpha * (centroid - p_worst)
	    f_ref = objective(p_reflect)
	
	    f_best = f_values[0]
	    f_second_worse = f_values[-2]
	    f_worst = f_values[-1]
	
	    # 反射点が一番良い場合
	    if f_ref < f_best:
	        # もうちょっと良い点がないか探りに行く (拡大点を試す)
	        p_expand = centroid + gamma * (centroid - p_worst)
	        f_expand = objective(p_expand)
	
	        # 拡大点が良かったら採用. だめだったら反射点で妥協
	        if f_expand < f_ref:
	            eval_points[-1] = p_expand
	        else:
	            eval_points[-1] = p_reflect
	
	    # 反射点がそこそこ良い場合 (second worse よりは良い) → 最悪点を反射点で置き換え
	    elif f_ref < f_second_worse:
	        eval_points[-1] = p_reflect
	    else: # f_ref >= f_second_worse
	        p_cont = centroid + rho * (p_reflect - centroid)
	        f_cont = objective(p_cont)
	        if f_cont < f_worst: # 縮小点を計算
	            # 縮小点が最悪点よりは良い時、最悪点を置き換え
	            eval_points[-1] = p_cont
	        else: # 最良点めがけて縮小させる
	            shrink_points =  eval_points[0] + sigma * (eval_points - eval_points[0])
	            eval_points[1:] = shrink_points[1:]
	    center = np.mean(eval_points, axis=0)

	# final check if the result is reasonable (Kenar)
	fvertex = objective(np.transpose(eval_points))
	(fcenter,fvmin) = (objective(center),np.min(fvertex))
	if fcenter - fvmin > err:
	    print("# Failed. not a local minimum:",fcenter,fvmin)
	return center

# Adaptive Nelder Mead Method by Gao and Han with inside contraction
# x0 is an initial value, a list of points
def anmsearch(objective, x0, err=1e-5,**k): # Kenar
	# a point is a list of n floats or n ints
	# x0 may be a point or a list of points 
	# the "list" may be a tuple,np.ndarray
	n = len(x0) - 1 	# dim
	if isinstance(x0[0],(list,tuple,np.ndarray)):
		eval_points = np.array(x0,dtype=float)		### Fix
	else:
		init_point =  np.array(x0,dtype=float)		### Fix
		scaling_factor = 1.
		if 'scale' in k:
			scaling_factor = k['scale']
		eval_points = np.vstack([init_point + np.eye(len(x0)) * scaling_factor, init_point])
	if n <=2:
		alpha = 1.0
		gamma = 2.0
		rho = 0.5
		sigma = 0.5
	else:							# let $n$ be len(x0)
		alpha = 1.0					# $\alpha$ by Gao, $\alpha=1$
		gamma = 1.0 + 2.0/n			# $\alpha\beta$  by Gao, $\beta=1+2/n$
		rho = .75 - 0.5/n			# $\gamma$ by Gao, $\gamma=3/4 - 1/(2n)$
		sigma = 1.0 - 1.0/n			# $\delta$ by Gao, $\delta=1-1/n$	
	center = np.mean(eval_points, axis=0)
	# loop while radius of the simplex is larger than err
	while np.max(abs(eval_points - center)) > err: ### Fix,Kenar
	    # 目的関数で評価して並び替え
	    objective_values_by_point = objective(np.transpose(eval_points)) # Kenar
	    orders = np.argsort(objective_values_by_point)
	    f_values = objective_values_by_point[orders]
	    eval_points = eval_points[orders]
	
	    # 最悪点以外で重心を計算
	    centroid = np.mean(eval_points[:-1], axis=0)
	    p_worst = eval_points[-1]
	
	    # 反射点 (最悪点から重心の対象な点) を計算 + そのときの目的関数を評価
	    p_reflect = centroid + alpha * (centroid - p_worst)
	    f_ref = objective(p_reflect)
	
	    f_best = f_values[0]
	    f_second_worse = f_values[-2]
	    f_worst = f_values[-1]
	
	    # 反射点が一番良い場合
	    if f_ref < f_best:
	        # もうちょっと良い点がないか探りに行く (拡大点を探す)
	        p_expand = centroid + gamma * (centroid - p_worst)
	        f_expand = objective(p_expand)
	
	        # 拡大点が良かったら採用. だめだったら反射点で妥協
	        if f_expand < f_ref:
	            eval_points[-1] = p_expand
	        else:
	            eval_points[-1] = p_reflect
	
	    # 反射点がそこそこ良い場合 (second worse よりは良い) → 最悪点を反射点で置き換え
	    elif f_ref < f_second_worse:
	        eval_points[-1] = p_reflect
	
	    else: # f_ref >= f_second_worse
	        renewed = False
	        if f_ref < f_worst: # 外側の縮小点を計算
	            p_cont = centroid + rho * (p_reflect - centroid)
	            f_cont = objective(p_cont)
	            if f_cont <= f_ref:
	                # 縮小点が反射点よりは良い時、最悪点を置き換え
	                eval_points[-1] = p_cont
	                renewed = True
	        else: # f_ref >= f_worst  # 内側の縮小点を計算
	            p_cont = centroid - rho * (p_reflect - centroid)
	            f_cont = objective(p_cont)
	            if f_cont < f_worst:
	                # 縮小点が最悪点よりは良い時、最悪点を置き換え
	                eval_points[-1] = p_cont
	                renewed = True
	        if not renewed:
	            # 最良点めがけて縮小させる
	            shrink_points =  eval_points[0] + sigma * (eval_points - eval_points[0])
	            eval_points[1:] = shrink_points[1:]
	    center = np.mean(eval_points, axis=0)
	# final check if the result is reasonable (Kenar)
	fvertex = objective(np.transpose(eval_points))
	(fcenter,fvmin) = (objective(center),np.min(fvertex))
	if fcenter - fvmin > err:
	    print("# Failed. not a local minimum:",fcenter,fvmin)
	return center
